import ElementBuilder from "./elementBuilder";
import { PieData, PieOption } from "./types";
import { randomRgbColor } from "./utils/color";
import { easeIn } from "./utils/ease";
import { Path } from "./utils/path";
import { pointOnCircle } from "./utils/point";
import { transition } from './utils/transition';

export default function pie(option: PieOption): SVGElement {
  const {
    width,
    height,
    data
  } = option;

  const svg = ElementBuilder.create('svg');
  svg.setAttrs({
    width, height,
    viewBox: '0 0 100 100',
  });

  svg.style('overflow', 'visible');

  const g = ElementBuilder.create('g');
  g.setAttrs({transform: 'translate(50, 50)'});
  svg.append(g);

  transition({
    from: 0,
    to: Math.PI * 2,
    ease: easeIn,
    duration: 500,
  }, (totalRad) => {
    const paths = render(data, totalRad as number);
    g.empty().append(...paths);
  });
  
  return svg.element;
}

function render(data: PieData = [], totalRad: number): ElementBuilder[] {
  let startRad = -Math.PI / 2;
  const total = data.reduce((sum, {value}) => sum += value, 0);
  const paths: ElementBuilder[] = [];
  data.forEach((item) => {
    if (!item.fillColor) {
      item.fillColor = randomRgbColor();
    }
    const {value, strokeColor, fillColor, strokeWidth = 0} = item;
    const fan = ElementBuilder.create('path');
    const path = new Path();
    path.moveTo(0, 0);

    const rad = value / total * totalRad;
    path.lineTo(...pointOnCircle(0, 0, 50, startRad));
    startRad += rad;
    path.ellipticalArcTo(
      50, 50, 0, rad > Math.PI, true, ...pointOnCircle(0, 0, 50, startRad)
    );
    path.closePath();

    fan.setAttrs({
      d: path.pathString,
      stroke: strokeColor || fillColor,
      fill: fillColor,
      'stroke-width': strokeWidth,
    });

    paths.push(fan);
  });

  return paths;

}