import { Instrument } from "src/types";

type PathSegment = [Instrument, ...number[]];

export class Path {
  private _pathSegments: PathSegment[];
  private _isClosed: boolean;
  constructor() {
    this._pathSegments = [];
    this._isClosed = false;
  }

  private addPathSegments(instrument: Instrument, ...params: number[]) {
    if (this._isClosed) {
      throw 'This path is already closed';
    }
    this._pathSegments.push([instrument, ...params]);
    if (instrument === 'Z') {
      this._isClosed = true;
    }
  }

  moveTo(x: number, y: number) {
    this.addPathSegments('M', x, y);
    return this;
  }

  moveToRelative(dx: number, dy: number) {
    this.addPathSegments('m', dx, dy);
    return this;
  }

  lineTo(x: number, y: number) {
    this.addPathSegments('L', x, y);
    return this;
  }

  lineToRelative(dx: number, dy: number) {
    this.addPathSegments('l', dx, dy);
    return this;
  }

  horizontalLineTo(x: number) {
    this.addPathSegments('H', x);
    return this;
  }

  horizontalLineToRelative(dx: number) {
    this.addPathSegments('h', dx);
    return this;
  }

  verticalLineTo(y: number) {
    this.addPathSegments('V', y);
    return this;
  }

  verticalLineToRelative(dy: number) {
    this.addPathSegments('v', dy);
    return this;
  }

  curveTo(cx1: number, cy1: number, cx2: number, cy2: number, tx: number, ty: number) {
    this.addPathSegments('C', cx1, cy1, cx2, cy2, tx, ty);
    return this;
  }

  curveToRelative(cdx1: number, cdy1: number, cdx2: number, cdy2: number, tdx: number, tdy: number) {
    this.addPathSegments('c', cdx1, cdy1, cdx2, cdy2, tdx, tdy);
    return this;
  }

  smoothCurveTo(cx2: number, cy2: number, tx: number, ty: number) {
    this.addPathSegments('S', cx2, cy2, tx, ty);
    return this;
  }

  smoothCurveToRelative(cdx2: number, cdy2: number, tdx: number, tdy: number) {
    this.addPathSegments('s', cdx2, cdy2, tdx, tdy);
    return this;
  }

  quadCurveTo(cx: number, cy: number, tx: number, ty: number) {
    this.addPathSegments('Q', cx, cy, tx, ty);
    return this;
  }

  quadCurveToRelative(cdx: number, cdy: number, tdx: number, tdy: number) {
    this.addPathSegments('q', cdx, cdy, tdx, tdy);
    return this;
  }

  smoothQuadCurveTo(tx: number, ty: number) {
    this.addPathSegments('T', tx, ty);
    return this;
  }

  smoothQuadCurveToRelative(tdx: number, tdy: number) {
    this.addPathSegments('t', tdx, tdy);
    return this;
  }

  ellipticalArcTo(
    rx: number,
    ry: number,
    xAxisRotation: number,
    isLargeArc: boolean,
    isClockWise: boolean,
    tx: number,
    ty: number
  ) {
    this.addPathSegments(
      'A',
      rx,
      ry,
      xAxisRotation,
      isLargeArc ? 1 : 0,
      isClockWise ? 1 : 0,
      tx,
      ty
    );
    return this;
  }

  closePath() {
    this.addPathSegments('Z');
    return this;
  }

  back() {
    this._pathSegments.pop();
    return this;
  }
  remove(index: number) {
    this._pathSegments.splice(index, 1);
    return this;
  }

  toString(): string {
    return this.pathString;
  }

  get pathString() {
    return this._pathSegments.map(([ins, ...params]: PathSegment) => {
      return ins + params.join(' ');
    }).join(' ');
  }

  append(path: Path): this {
    if (this._isClosed) {
      throw 'This path is already closed';
    }
    this._pathSegments = [...this._pathSegments, ...path._pathSegments];
    if (this._pathSegments[this._pathSegments.length - 1][0] === 'Z') {
      this._isClosed = true;
    }
    return this;
  }
}